<?php
/**
 * Page Output
 * @package     Page
 * @copyright   Copyright (C) 2011 MicroEngine Inc.
 */
class Page_Output extends Page
{
    /** キャッシュディレクトリ */
    const CACHE_DIR = '/cache';

    /**
     * @var Page_Dao
     */
    private $dao;
    /**
     * @var Metag_Dao
     */
    private $metag_dao;
    /**
     * @var Template_Logic
     */
    private $template_logic;
    /**
     * @var Template_Parser
     */
    private $parser;
    /**
     * Document
     * @var Metag_Template
     */
    private $document;
    /**
     * Page内のmetagのデータ
     * @var array
     */
    private $data;
    /**
     * 出力ファイル
     * @var string
     */
    private $output_file;
    /**
     * 先頭スクリプト
     * @var string
     */
    private $prepend_script;
    /**
     * モジュール用出力ファイル
     * @var array
     */
    private $module_output_list;
    /**
     * 参照ページデータ
     * @var array
     */
    private $ref_page_data;
    /**
     * 参照テンプレートデータ
     * @var array
     */
    private $ref_template_repeat;
    /**
     * @var Template_Form;
     */
    private $template_form;

    /**
     * コンストラクタ
     */
    public function __construct()
    {
        $this->dao = new Page_Dao();
        $this->metag_dao = new Metag_Dao();
        $this->template_logic = new Template_Logic();
        $this->parser = new Template_Parser();
    }

    /**
     * ページ書き出し
     * @param string $page_id
     * @return boolean
     */
    public function output($page_id)
    {
        // data初期化
        $this->data = array();

        // ページ情報取得
        $this->_id = $page_id;
        if (!$this->dao->set_page($this)) {
            return false;
        }

        if ($this->page_type !== Page_Logic::PAGE_TYPE_TEMPLATE) {
            return false;
        }

        // テンプレート情報取得
        // template_idが同じ場合は再利用する
        if ($this->template_form == null || $this->template_form->_id !== $this->template_id) {
            $this->template_form = $this->template_logic->get_template_form_by_page($this);

            // ページ情報からテンプレートを作成
            $this->template_logic->get_template_file($this->template_form);
            // parse処理
            $this->parser->output_method = $this->template_form->output_method;
            $this->document = $this->parser->parse($this->template_form->source_code, $this->page_path);
        }

        if ($this->public_flag == 0) {
            // 非公開の場合はファイルを削除する
            $file_path = SITE_ROOT . $this->page_path;
            if (file_exists($file_path)) {
                unlink($file_path);
            }
            if ($this->template_form->template_type === Template_Logic::TYPE_ARCHIVE || $this->template_form->template_type === Template_Logic::TYPE_PAGINATION) {
                // 関連ページネーション・アーカイブページを削除する
                $page_logic = new Page_Logic();
                $page_logic->delete_output_page($this->template_form->_id);
            }
            return true;
        }

        // ページデータ取得
        $this->set_data($this->metag_dao->get_page_data($this->_id));
        // ページ基本情報セット
        $this->set_basic_info();

        $this->output_file = '';
        $this->prepend_script = '';
        if ($this->template_form->output_method === Template_Logic::OUTPUT_DYNAMIC) {
            // 先頭スクリプト追加
            $file_path = SITE_PATH . $this->page_path;
            // site.phpの絶対パス
            $me_path = CMS_PATH . Template_Logic::SITE_FILE_PATH;
            // index.phpへの相対パスを取得
            $relative_path = ME_Util::get_relative_path($file_path, $me_path, true);
            $this->prepend_script = "<?php include_once('" . $relative_path . "') ?>";
            $this->output_file = $this->prepend_script;
        }

        // 参照データを取得する
        $this->ref_page_data = array();
        foreach ($this->document->ref_page_list as $ref_page_id) {
            $rows = $this->metag_dao->select_by_page_id($ref_page_id, true);
            // idを添字にした配列に変換
            $data = $this->convert_assoc_by_id($rows);
            $this->ref_page_data[$ref_page_id] = $data;
        }
        foreach ($this->document->ref_template_repeat as $repeat) {
            $rows = $this->metag_dao->select_by_template_id($repeat->ref_template_id, $repeat->limit, $repeat->offset, $repeat->sort_by, $repeat->sort_order);
            $this->data[$repeat->id] = $rows;
        }

        if ($this->template_form->template_type === Template_Logic::TYPE_PAGINATION) {
            // ページネーションタイプの場合
            $this->output_pagination_page();
        } else if ($this->template_form->template_type === Template_Logic::TYPE_ARCHIVE) {
            // アーカイブタイプの場合
            $this->output_archive_page();
        } else {
            $this->output_page($this->page_path);
        }
    }

    /**
     * プレビュー
     * @param Page_Form $form
     */
    public function preview(Page_Form $form)
    {
        // ページ情報取得
        $this->_id = $form->_id;
        $this->template_id = $form->template_id;
        $this->page_type = $form->page_type;
        $this->page_path = $form->page_path;
        $this->title = $form->title;
        $this->keywords = $form->keywords;
        $this->description = $form->description;
        $this->create_date = $form->create_date;
        $this->update_date = $form->update_date;

        // テンプレート情報取得
        $this->template_form = $form->template_form;
        // ソースコード再取得
        $this->template_logic->get_template_file($this->template_form);
        $this->template_form->source_code = $this->template_logic->add_base_tag(
            $this->template_form->source_code, $this->template_form->template_path);
        // parse処理
        $this->parser->output_method = Template_Logic::OUTPUT_STATIC;
        $this->document = $this->parser->parse($this->template_form->source_code, $this->page_path);

        // ページデータセット。空のリピート領域を削除する。
        $this->set_data($this->delete_empty_repeat($form));
        // ページ基本情報セット
        $this->set_basic_info();

        $this->output_file = '';

        // 参照データを取得する
        $this->ref_page_data = array();
        foreach ($this->document->ref_page_list as $ref_page_id) {
            $rows = $this->metag_dao->select_by_page_id($ref_page_id, true);
            // idを添字にした配列に変換
            $data = $this->convert_assoc_by_id($rows);
            $this->ref_page_data[$ref_page_id] = $data;
        }
        foreach ($this->document->ref_template_repeat as $repeat) {
            $rows = $this->metag_dao->select_by_template_id($repeat->ref_template_id, $repeat->limit, $repeat->offset, $repeat->sort_by, $repeat->sort_order);
            $this->data[$repeat->id] = $rows;
        }

        if ($this->template_form->template_type === Template_Logic::TYPE_PAGINATION) {
            // アーカイブタイプの場合は現在のリストに絞り込む
            $this->output_pagination_page(false);
        } else if ($this->template_form->template_type === Template_Logic::TYPE_ARCHIVE) {
            // アーカイブタイプの場合は現在のリストに絞り込む
            $this->output_archive_page(false);
        }

        $this->composition($this->document);

        if ($this->template_form->char_code !== null && $this->template_form->char_code !== Template_Logic::CHAR_CODE) {
            $this->output_file = mb_convert_encoding($this->output_file, $this->template_form->char_code, Template_Logic::CHAR_CODE);
        }

        // ファイル出力
        h::no_escape($this->output_file);
    }

    /**
     * キャッシュファイル書き出し
     * @param array $sub_template_list
     */
    private function output_cache($sub_template_list)
    {
        $cache_list = array();
        // メインテンプレートのキャッシュファイル書き出し
        Me_Util::write_file(DATA_ROOT . self::CACHE_DIR . $this->page_path, $this->output_file);
        foreach ($sub_template_list as $sub_template) {
            $sub_template_form = $this->template_logic->select($sub_template['_id'], 'Template_Form');

            // ページ情報からテンプレートを作成
            $this->template_logic->get_template_file($sub_template_form);
            // parse処理
            $this->parser->output_method = $sub_template_form->output_method;
            $sub_template_document = $this->parser->parse($sub_template_form->source_code, $this->page_path);

            $this->output_file = '';
            if ($sub_template_form->output_method === Template_Logic::OUTPUT_DYNAMIC) {
                // 先頭スクリプト追加
                $file_path = SITE_PATH . $this->page_path;
                // site.phpの絶対パス
                $me_path = CMS_PATH . Template_Logic::SITE_FILE_PATH;
                // index.phpへの相対パスを取得
                $relative_path = ME_Util::get_relative_path($file_path, $me_path, true);
                $this->prepend_script = "<?php include_once('" . $relative_path . "') ?>";
                $this->output_file = $this->prepend_script;
            }

            // 参照データを取得する
            foreach ($sub_template_document->ref_page_list as $ref_page_id) {
                // メインテンプレートで取得していない場合だけ、データを取得する
                if (!isset($this->ref_page_data[$ref_page_id])) {
                    $rows = $this->metag_dao->select_by_page_id($ref_page_id, true);
                    // idを添字にした配列に変換
                    $data = $this->convert_assoc_by_id($rows);
                    $this->ref_page_data[$ref_page_id] = $data;
                }
            }
            foreach ($sub_template_document->ref_template_repeat as $repeat) {
                // limit, offsetなど、メインテンプレートと指定が違う可能性があるので、取得し直す。
                $rows = $this->metag_dao->select_by_template_id($repeat->ref_template_id, $repeat->limit, $repeat->offset, $repeat->sort_by, $repeat->sort_order);
                $this->data[$repeat->id] = $rows;
            }

            $this->module_output_list = array();
            $this->composition($sub_template_document);

            if ($sub_template_form->char_code !== null && $sub_template_form->char_code !== Template_Logic::CHAR_CODE) {
                $this->output_file = mb_convert_encoding($this->output_file, $sub_template_form->char_code, Template_Logic::CHAR_CODE);
            }
            // サブテンプレートのキャッシュファイル書き出し
            Me_Util::write_file(DATA_ROOT . self::CACHE_DIR . $this->page_path . '.' . basename($sub_template_form->template_path), $this->output_file);
            $cache_list[$sub_template['cond']] = self::CACHE_DIR . $this->page_path . '.' . basename($sub_template_form->template_path);

            // モジュール書き出し
            foreach ($this->module_output_list as $module_name => $module_file) {
                // ファイル名のチェック
                if (!preg_match('|^[\w-./]+$|', $module_name)) {
                    // a-zA-Z-9_-./ 以外の文字があれば書き出さない
                    continue;
                }
                if ($sub_template_form->char_code !== null && $sub_template_form->char_code !== Template_Logic::CHAR_CODE) {
                    $module_file = mb_convert_encoding($module_file, $sub_template_form->char_code, Template_Logic::CHAR_CODE);
                }
                // ファイル書き出し
                Me_Util::write_file(DATA_ROOT . '/module/' . $module_name, $module_file);
            }
        }
        $cache_list['_DEFAULT'] = self::CACHE_DIR . $this->page_path;

        // 公開側ページの書き出し
        // 先頭スクリプト追加
        $file_path = SITE_PATH . $this->page_path;
        // site.phpの絶対パス
        $me_path = CMS_PATH . Template_Logic::SITE_FILE_PATH;
        // index.phpへの相対パスを取得
        $relative_path = ME_Util::get_relative_path($file_path, $me_path, true);
        $output_file = "<?php include_once('" . $relative_path . "');\n"
            . '$cache_list = ' . var_export($cache_list, true) . ";\n"
            . '$cache_path = $helper->get_cache_path($cache_list);' . "\n"
            . 'include($cache_path); ?>';
        $result = Me_Util::write_file(SITE_ROOT . $this->page_path, $output_file);
        $this->output_page_result($this->page_path, $result);
    }

    /**
     * 出力ファイル組み立て
     * @param object $node Metag_*
     * @param string $repeat_no
     * @param string $module_name
     */
    private function composition($node, $repeat_no = null, $module_name = '')
    {
        if ($node->node_type === 'module' && !empty($node->module_name)) {
            $module_name = $node->module_name;
        }

        // block要素のみ、子を呼び出す
        if ($node->is_block) {
            switch ($node->node_type) {
                case 'repeat':
                    $i = 1;
                    $total = $count = $this->count_repeat_rows($node);
                    if (is_numeric($node->offset)) {
                        $i += (integer)$node->offset;
                    }
                    if (is_numeric($node->limit)) {
                        if ($count > $node->limit) {
                            $count = $node->limit;
                        }
                    }
                    // 等価確認のため、リピート領域内のデータを取得
                    $repeat_rows = $this->get_row($node);

                    // 並べ替え
                    $this->sort_repeat_rows($node, $repeat_rows);

                    // 表示カウント
                    $display_count = 1;
                    for (; $i <= $total; $i++) {
                        $row_number = (isset($repeat_rows[$i]['repeat_no_orig'])) ? $repeat_rows[$i]['repeat_no_orig'] : $i;
                        $repeat_row = $repeat_rows[$i];
                        if (!$this->equal($node, $repeat_row)) {
                            // equal属性が指定されていて、同じではない場合は、書き出さない。
                            continue;
                        }
                        // equalで絞込んだ場合に、実際の表示数でlimitするための処理
                        if ($display_count <= $count) {
                            $display_count++;
                        } else {
                            break;
                        }
                        foreach ($node->child_nodes as $child) {
                            $child->repeat_count = $count;
                            $child->display_count = $display_count - 1;
                            $this->composition($child, $row_number, $module_name);
                        }
                    }
                    break;
                case 'delete';
                    $exist_child_value = false;
                    $this->exist_child_value($node, $repeat_no, $exist_child_value);
                    if ($exist_child_value) {
                        foreach ($node->child_nodes as $child) {
                            $this->composition($child, $repeat_no, $module_name);
                        }
                    }
                    break;
                case 'display';
                    if ($node->site !== 'off') {
                        foreach ($node->child_nodes as $child) {
                            $this->composition($child, $repeat_no, $module_name);
                        }
                    }
                    break;
                default:
                    foreach ($node->child_nodes as $child) {
                        $this->composition($child, $repeat_no, $module_name);
                    }
                    break;
            }
        } else {
            // 該当ノードのデータを取り出す
            $row = $this->get_row($node, $repeat_no);

            if (!empty($module_name)) {
                $output = $node->output($row);
                $this->output_file .= $output;
                $this->module_output_list[$module_name] .= $output;
            } else {
                $this->output_file .= $node->output($row);
            }
        }

        if ($node->node_type === 'module') {
            $module_name = '';
        }
    }

    /**
     * MEタグデータ取得
     * @param object $node
     * @param string $repeat_no
     * @param boolean $check_exist_flag 存在確認フラグ
     * @return array
     */
    private function get_row($node, $repeat_no = null, $check_exist_flag = false)
    {
        $ret = null;
        // htmlノードはnullを返す。
        if ($node->node_type === 'html') {
            return $ret;
        }

        $id = (empty($node->ref_id)) ? $node->id : $node->ref_id;

        if (strlen($node->ref_page_id) > 0) {
            $data = $this->ref_page_data[$node->ref_page_id];
        } else {
            $data = $this->data;
        }

        if ($id === null || empty($node->repeat_id) || empty($repeat_no)) {
            if (isset($id) && isset($data[$id])) {
                $ret = $data[$id];
            }
        } else {
            // repeat
            $repeat_id = (empty($node->ref_repeat_id)) ? $node->repeat_id : $node->ref_repeat_id;
            if (isset($data[$repeat_id][$repeat_no][$id])) {
                $ret = $data[$repeat_id][$repeat_no][$id];
            }
        }

        // リピート領域内の場合は、空でもrepeat_noとpage_idだけは返す
        if ($ret === null && $repeat_no !== null && $check_exist_flag === false) {
            $ret = array('repeat_no' => $repeat_no, '_page_id' => $data[$repeat_id][$repeat_no]['_page_id']['node_value']);
            if (strlen($node->initial_value)) {
                $ret['node_value'] = $node->initial_value;
            }
        }

        // 処理中のページIDを追加
        $ret['_self_page_id'] = $this->_id;

        return $ret;
    }

    /**
     * repeat内データの有無確認
     * @param object $node
     * @return integer
     */
    private function count_repeat_rows($node)
    {
        if ($node->ref_page_id !== null) {
            $data = $this->ref_page_data[$node->ref_page_id];
        } else {
            $data = $this->data;
        }
        $id = (empty($node->ref_id)) ? $node->id : $node->ref_id;

        return count($data[$id]);
    }

    /**
     * idを添字にした配列に変換
     * @param array $array_list
     * @return array $assoc
     */
    private function convert_assoc_by_id($array_list)
    {
        $assoc = array();
        foreach ($array_list as $array) {
            // repeatタグの親子関係を整理
            if (!empty($array['repeat_id'])) {
                $assoc[$array['repeat_id']][$array['repeat_no']][$array['id']] = $array;
            } else {
                $assoc[$array['id']] = $array;
            }
        }
        return $assoc;
    }

    /**
     * ページ基本情報セット
     */
    private function set_basic_info()
    {
        if (!is_array($this->data)) {
            $this->data = array();
        }
        $mapping = array(
            '_page_id' => '_id',
            '_template_id' => 'template_id',
            '_path' => 'page_path',
            '_title' => 'title',
            '_keywords' => 'keywords',
            '_description' => 'description',
        );
        foreach ($mapping as $key => $prop) {
            $row = array(
                'node_value' => $this->$prop
            );
            $this->data[$key] = $row;
        }
        $this->data['_create_date'] = array('node_value' => date(Me_Config::SYSTEM_DATE_FORMAT, $this->create_date));
        $this->data['_update_date'] = array('node_value' => date(Me_Config::SYSTEM_DATE_FORMAT, $this->update_date));
        $this->data['_now'] = array('node_value' => NOW);
    }

    /**
     * $dataプロパティに配列を追加する
     * @param array $compact_data
     */
    private function set_data($compact_data)
    {
        if (!is_array($this->data)) {
            $this->data = array();
        }
        $this->data = array_merge($this->data, $compact_data);
    }

    /**
     * 空のリピート領域削除
     * @param Page_Form $form
     * @return array $data
     */
    private function delete_empty_repeat(Page_Form $form)
    {
        $data = array();
        foreach ($form->data as $key => $row) {
            $metag = $form->document->get_element_by_id($key);
            if ($metag === null || $metag->repeat_id != $row['repeat_id'] || strlen($metag->ref_id) > 0 || substr($metag->id, 0, 1) === '_') {
                // repeat_idが一致しない場合（空同士一致でも良い）はpreviewから除外
                // あるいは、参照元の場合。
                continue;
            }
            if ($metag->node_type === 'repeat') {
                $data[$key] = array();
                // inside of repeat tag
                $new_repeat_no = 1;
                foreach ($row as $repeat_no => $repeat_list) {
                    $exist = false;
                    foreach ($repeat_list as $repeat_inside_id => $repeat_inside_row) {
                        $repeat_inside_metag = $form->document->get_element_by_id($repeat_inside_id);
                        $repeat_inside_row['repeat_no'] = $new_repeat_no;
                        $repeat_inside_row['page_id'] = $form->_id;
                        // repeat内は値が無い場合は除外
                        // node_valueは配列の場合がある
                        if ((is_array($repeat_inside_row['node_value']) && count($repeat_inside_row['node_value']) > 0)
                            || (!is_array($repeat_inside_row['node_value']) && strlen($repeat_inside_row['node_value']) > 0)) {
                            $data[$key][$new_repeat_no][$repeat_inside_id] = $repeat_inside_row;
                            $exist = true;
                        }
                    }
                    if ($exist) {
                        $new_repeat_no++;
                    }
                }
            } else {
                $row['id'] = $key;
                $row['page_id'] = $form->_id;
                $data[$key] = $row;
            }
        }
        return $data;
    }

    /**
     * ページネーションページ書き出し
     * @param boolean $output_mode
     */
    private function output_pagination_page ($output_mode = true)
    {
        // テンプレート設定はアーカイブと同じカラムを使用する。
        $target_repeat_id = $this->template_form->archive_repeat_id;
        $prefix = $this->template_form->archive_prefix;
        $suffix = $this->template_form->archive_suffix;
        $per_page = $this->template_form->per_page;
        if (empty($per_page)) {
            $per_page = 10;
        }

        // 対象リピートタグを取得
        $repeat_tag = $this->document->get_element_by_id($target_repeat_id);

        // ページネーションデータリストを取得
        $pagination_data_list = $this->data[$target_repeat_id];
        // 標準日付カラム以外での並べ替えを実行する。
        $this->sort_repeat_rows($repeat_tag, $pagination_data_list);

        // 同じディレクトリ内の既存ページを取得
        $exist_page_list_rows = $this->dao->select_page_list(dirname($this->page_path));
        $exist_page_list = array();
        $exist_pagination_page = array();
        // page_pathをキーにする
        foreach ($exist_page_list_rows as $exist_page) {
            // ページパスをキーにした配列を作成
            $exist_page_list[$exist_page['page_path']] = $exist_page;
            // ページネーションテンプレートを元に書き出されたページネーションページのリストを作成
            if ($this->template_form->_id === $exist_page['template_id'] && $exist_page['page_type'] == Page_Logic::PAGE_TYPE_TEMPLATE_OUTPUT) {
                $exist_pagination_page[$exist_page['page_path']] = $exist_page;
            }
        }

        $paginator = new Page_Paginator($pagination_data_list, $this->page_path, $prefix, $suffix, $per_page);

        // プレビューモードの場合
        if (!$output_mode) {
            $this->data[$target_repeat_id] = $paginator->current_items();
            $this->data = array_replace($this->data, $paginator->get_reserved());
            return;
        }

        // 各ページを書き出す
        do {
            $page_path = $paginator->get_page_path();

            // 新ページがテンプレート化されていないファイルだったらテーブルから削除する。
            if (isset($exist_page_list[$page_path]) && $exist_page_list[$page_path]['page_type'] == Page_Logic::PAGE_TYPE_FILE) {
                $this->dao->destroy($exist_page_list[$page_path]['_id']);
                unset($exist_page_list[$page_path]);
            }

            // 既存ページがなければページネーションページとして登録
            if (!isset($exist_page_list[$page_path])) {
                $this->register_output_page($page_path);
            }
            // 既存アーカイブページでこの時点で存在するものを省く
            unset($exist_pagination_page[$page_path]);

            $this->data[$target_repeat_id] = $paginator->current_items();
            $this->data = array_replace($this->data, $paginator->get_reserved());
            $this->output_page($page_path);
        } while ($paginator->page_feed());

        // 不要な既存ページネーションページを削除
        foreach ($exist_pagination_page as $pagination_page) {
            if (file_exists(SITE_ROOT . $pagination_page['page_path'])) {
                unlink(SITE_ROOT . $pagination_page['page_path']);
            }
            $this->dao->destroy($pagination_page['_id']);
        }
    }

    /**
     * アーカイブページ書き出し
     * @param boolean $output_mode
     */
    private function output_archive_page ($output_mode = true)
    {
        $archive_repeat_id = $this->template_form->archive_repeat_id;
        $archive_start_month = (strlen($this->template_form->archive_start_month) > 0) ? $this->template_form->archive_start_month : '01';
        $archive_start_day = (strlen($this->template_form->archive_start_day) > 0) ? $this->template_form->archive_start_day : '01';
        $archive_start = '/' . $archive_start_month . '/' . $archive_start_day;
        $archive_prefix = $this->template_form->archive_prefix;
        $archive_suffix = $this->template_form->archive_suffix;

        $repeat_tag = $this->document->get_element_by_id($archive_repeat_id);
        // v1.3.0 ではtemplateテーブルのarchive_date_idを使っていたが、v1.4.0からリピートタグのsort_byを使うように変更した。
        $sort_by = (strlen($repeat_tag->sort_by) > 0) ? $repeat_tag->sort_by : '_create_date';

        // 標準日付カラム以外での並べ替えを実行する。
        $repeat_rows = $this->data[$archive_repeat_id];
        $this->sort_repeat_rows($repeat_tag, $repeat_rows);

        // アーカイブデータリストを作成
        $archive_data_list = array();
        foreach ($repeat_rows as $archive_item) {
            if (!$this->equal($repeat_tag, $archive_item)) {
                continue;
            }

            $archive_name = date('Y', strtotime($archive_item[$sort_by]['node_value']));
            // 開始日を元に、年を調整
            if (strtotime($archive_item[$sort_by]['node_value']) < strtotime($archive_name . $archive_start)) {
                $archive_name--;
            }
            // repeatリストは1から開始させる。
            $no = count($archive_data_list[$archive_name]) + 1;
            $archive_data_list[$archive_name][$no] = $archive_item;
        }

        // 今の期間を定義
        $current_archive_name = date('Y');
        // 開始日を元に、年を調整
        if (time() < strtotime($current_archive_name . $archive_start)) {
             $current_archive_name--;
        }

        // 今の期間に該当するデータがない場合は、空配列を先頭に追加
        if (!isset($archive_data_list[$current_archive_name])) {
            $archive_data_list_keys = array_keys($archive_data_list);
            if (!empty($archive_data_list_keys)) {
                $current_archive_name = $archive_data_list_keys[0];
            } else {
                $archive_data_list[$current_archive_name] = array();
            }
        }

        // アーカイブリスト（リンク用）を作成
        $archive_list = array();
        $register_archive_list = array();
        foreach ($archive_data_list as $archive_name => $archive_items) {
            $no = count($archive_list) + 1;
            // 現在の期間は元のpage_pathを使う
            if ((string) $current_archive_name === (string) $archive_name) {
                $archive_list[$no]['_path']['node_value'] = SITE_PATH . $this->page_path;
                $register_archive_list[$no]['_archive_name']['ref_link_path'] = $this->page_path;
            } else {
                $archive_list[$no]['_path']['node_value'] = SITE_PATH . $this->get_archive_path($this->page_path, $archive_name, $archive_prefix, $archive_suffix);
                $register_archive_list[$no]['_archive_name']['ref_link_path'] = $this->get_archive_path($this->page_path, $archive_name, $archive_prefix, $archive_suffix);
            }
            $archive_list[$no]['_archive_name']['node_value'] = $archive_name;
            $archive_list[$no]['_archive_name']['ref_link_path'] = $archive_list[$no]['_path']['node_value'];
        }
        $this->data['_archive_list'] = $archive_list;

        if (!$output_mode) {
            // プレビューモード
            // 該当期間だけのデータをセットする
            $this->data[$archive_repeat_id] = $archive_data_list[$current_archive_name];
            $this->data['_archive_name']['node_value'] = $current_archive_name;
            return;
        }

        // アーカイブページを登録・削除
        $this->register_archive_page($register_archive_list);

        foreach ($archive_data_list as $archive_name => $archive_items) {
            // 該当期間だけのデータをセットする
            $this->data[$archive_repeat_id] = $archive_items;

            $page_path = $this->page_path;
            $this->data['_archive_name']['node_value'] = '';
            // 期間名を予約IDとして追加
            $this->data['_archive_name']['node_value'] = $archive_name;
            if ((string) $current_archive_name !== (string) $archive_name) {
                // アーカイブページ名で書きだす
                $page_path = $this->get_archive_path($this->page_path, $archive_name, $archive_prefix, $archive_suffix);
            }
            $this->output_page($page_path);
        }
    }

    /**
     * ページ・モジュール書き出し
     * @param string $page_path
     */
    private function output_page ($page_path)
    {
        $this->module_output_list = array();
        $this->output_file = $this->prepend_script;
        $this->composition($this->document);

        if ($this->template_form->char_code !== null && $this->template_form->char_code !== Template_Logic::CHAR_CODE) {
            $this->output_file = mb_convert_encoding($this->output_file, $this->template_form->char_code, Template_Logic::CHAR_CODE);
        }

        // モジュール書き出し
        foreach ($this->module_output_list as $module_name => $module_file) {
            // アーカイブページのモジュールは書き出さない。
            if ($page_path != $this->page_path) {
                continue;
            }
            // ファイル名のチェック
            if (!preg_match('|^[\w-./]+$|', $module_name)) {
                // a-zA-Z-9_-./ 以外の文字があれば書き出さない
                continue;
            }
            if ($this->template_form->char_code !== null && $this->template_form->char_code !== Template_Logic::CHAR_CODE) {
                $module_file = mb_convert_encoding($module_file, $this->template_form->char_code, Template_Logic::CHAR_CODE);
            }
            // ファイル書き出し
            Me_Util::write_file(DATA_ROOT . '/module/' . $module_name, $module_file);
        }

        // メインテンプレートと指定しているサブテンプレートを取得
        $sub_template_list = $this->template_logic->select_sub_template_list($this->template_form->_id);
        if (count($sub_template_list) == 0) {
            // サブテンプレートを持たない場合
            // 公開側ファイル書き出し
            $result = Me_Util::write_file(SITE_ROOT . $page_path, $this->output_file);
            $this->output_page_result($page_path, $result);
        } else {
            // サブテンプレートを持つ、メインテンプレートの場合
            $this->output_cache($sub_template_list);
        }
    }

    /**
     * ページ書き出し結果の保持とログ出力
     * @param string $path
     * @param integer $result
     */
    private function output_page_result ($path, $result)
    {
        Me_Util::$write_file_list[] = array('path' => $path, 'result' => $result);

        // 書き出し成功時はログを出力する。
        if ($result === Me_Util::WRITE_FILE_SUCCESS) {
            Me_Logger::log(1604, array($path));
        }
    }

    /**
     * アーカイブパス取得
     * @param string $page_path
     * @param string $archive_name
     * @param string $prefix
     * @param string $suffix
     * @return string
     */
    private function get_archive_path($page_path, $archive_name, $prefix = '', $suffix = '')
    {
        return rtrim(dirname($page_path), '/') . '/' . $prefix . $archive_name . $suffix . '.' . pathinfo($page_path, PATHINFO_EXTENSION);
    }

    /**
     * アーカイブページの登録・削除
     * @param array $archive_list
     */
    private function register_archive_page($archive_list)
    {
        // 同じディレクトリ内の既存ページを取得
        $exist_page_list = $this->dao->select_page_list(dirname($this->page_path));
        $exist_archive_page = array();
        // page_pathをキーにする
        foreach ($exist_page_list as $key => $exist_page) {
            // ページパスをキーにした配列を作成
            $exist_page_list[$exist_page['page_path']] = $exist_page;
            // アーカイブタイプテンプレートを元に書き出されたアーカイブページのリストを作成
            if ($this->template_form->_id === $exist_page['template_id'] && $exist_page['page_type'] == Page_Logic::PAGE_TYPE_TEMPLATE_OUTPUT) {
                $exist_archive_page[$exist_page['page_path']] = $exist_page;
            }
            unset($exist_page_list[$key]);
        }

        // アーカイブページをDBに登録
        foreach ($archive_list as $archive_item) {
            $new_page_path = $archive_item['_archive_name']['ref_link_path'];

            // 新ページがテンプレート化されていないファイルだったらテーブルから削除する。
            if (isset($exist_page_list[$new_page_path]) && $exist_page_list[$new_page_path]['page_type'] == Page_Logic::PAGE_TYPE_FILE) {
                $this->dao->destroy($exist_page_list[$new_page_path]['_id']);
                unset($exist_page_list[$new_page_path]);
            }

            // 既存ページがなければアーカイブページとして登録
            if (!isset($exist_page_list[$new_page_path])) {
                $this->register_output_page($new_page_path);
            }

            // 既存アーカイブページでこの時点で存在するものを省く
            unset($exist_archive_page[$new_page_path]);
        }

        // 不要な既存アーカイブページを削除
        foreach ($exist_archive_page as $archive_page) {
            if (file_exists(SITE_ROOT . $archive_page['page_path'])) {
                unlink(SITE_ROOT . $archive_page['page_path']);
            }
            $this->dao->destroy($archive_page['_id']);
        }
    }

    /**
     * テンプレート出力ページを登録
     * @param string $page_path
     */
    private function register_output_page($page_path)
    {
        $path_parts = pathinfo($page_path);
        $file_name = substr($path_parts['basename'], 0, strlen($path_parts['basename']) - strlen($path_parts['extension']) - 1);

        // pageオブジェクト作成
        $page = new Page();
        $page->template_id = $this->template_form->_id;
        $page->directory = rtrim(dirname($page_path), '/') . '/';
        $page->extension = '.' . $path_parts['extension'];
        $page->file_name = $file_name;
        $page->page_path = $page_path;
        $page->page_type = Page_Logic::PAGE_TYPE_TEMPLATE_OUTPUT;
        $this->dao->insert($page);
    }

    /**
     * 等価確認
     * @param object $node
     * @param array $repeat_row
     * @return boolean $result
     */
    private function equal ($node, $repeat_row)
    {
        $result = true;
        if (strlen($node->equal) > 0) {
            list($equal_id, $equal_value) = array_map('trim', explode(':', $node->equal));
            if ($repeat_row[$equal_id]['node_value'] != $equal_value) {
                $result = false;
            }
        }
        return $result;
    }

    /**
     * delete内の子ノードの値の有無を確認する
     * @param Object $node
     * @param string $repeat_no
     * @param boolean $exist
     */
    private function exist_child_value ($node, $repeat_no, &$exist)
    {
        // 基本は子ノードを対象にする
        $target_nodes = $node->child_nodes;

        // target_id が指定されている場合は、その属性に指定されたIDだけを対象にする
        if (strlen($node->target_id) > 0) {
            $target_nodes = array();
            $ids = Me_Util::get_array($node->target_id);
            foreach($ids as $id) {
                $target_nodes[] = $this->document->get_element_by_id($id);
            }
        }

        foreach ($target_nodes as $child) {
            // htmlではなく、通常のMEタグの場合
            if (strlen($child->id) > 0) {
                $child_row = $this->get_row($child, $repeat_no, true);
                if ($child->exist_node_value($child_row)) {
                    $exist = true;
                    break;
                }
            }
            // repeatの場合
            if ($child->node_type === 'repeat') {
                $i = 1;
                $total = $count = $this->count_repeat_rows($child);
                if (is_numeric($child->offset)) {
                    $i += (integer)$child->offset;
                }
                if (is_numeric($child->limit)) {
                    if ($count > $child->limit) {
                        $count = $child->limit;
                    }
                }
                // 等価確認のため、リピート領域内のデータを取得
                $repeat_rows = $this->get_row($child);

                // 並べ替え
                $this->sort_repeat_rows($node, $repeat_rows);

                // 表示カウント
                $display_count = 1;
                for (; $i <= $total; $i++) {
                    $row_number = (isset($repeat_rows[$i]['repeat_no_orig'])) ? $repeat_rows[$i]['repeat_no_orig'] : $i;
                    $repeat_row = $repeat_rows[$i];
                    if (!$this->equal($child, $repeat_row)) {
                        // equal属性が指定されていて、同じではない場合は、書き出さない。
                        continue;
                    }
                    // equalで絞込んだ場合に、実際の表示数でlimitするための処理
                    if ($display_count <= $count) {
                        $display_count++;
                    } else {
                        break;
                    }
                    foreach ($child->child_nodes as $repeat_child) {
                        $repeat_child->repeat_count = $count;
                        $repeat_child->display_count = $display_count - 1;
                        //$this->composition($repeat_child, $i, $module_name);
                        if (strlen($repeat_child->id) > 0) {
                            $repeat_child_row = $this->get_row($repeat_child, $row_number, true);
                            if ($child->exist_node_value($repeat_child_row)) {
                                $exist = true;
                                break;
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * リピート内の並べ替え
     * @param object $node Metag_*
     * @param array &$repeat_rows
     */
    private function sort_repeat_rows($node, &$repeat_rows)
    {
        // DBカラム以外での並べ替え
        if (strlen($node->sort_by) > 0
            && $node->sort_by !== Me_Sql::CREATE_DATE && $node->sort_by !== '_' . Me_Sql::CREATE_DATE
            && $node->sort_by !== Me_Sql::UPDATE_DATE && $node->sort_by !== '_' . Me_Sql::UPDATE_DATE
        ) {
            $repeat_rows_temp = $repeat_rows;
            $node_value = array();
            $_self_page_id = $repeat_rows_temp['_self_page_id'];
            unset($repeat_rows_temp['_self_page_id']);
            foreach($repeat_rows_temp as $key => $row) {
                    $node_value[$key] = $row[$node->sort_by]['node_value'];
                    $repeat_rows_temp[$key]['repeat_no_orig'] = $key;
            }
            $sort_order = ($node->sort_order === Me_Sql::ASC) ? SORT_ASC : SORT_DESC;
            array_multisort($node_value, $sort_order, $repeat_rows_temp);
            $repeat_rows = array();
            $j = 1;
            foreach ($repeat_rows_temp as $row) {
                $repeat_rows[$j] = $row;
                $j++;
            }
            if ($_self_page_id !== null) {
                $repeat_rows['_self_page_id'] = $_self_page_id;
            }
        }
    }
}