<?php
/**
 * Static Paginator
 * @package     Page
 * @copyright   Copyright (C) 2017 MicroEngine Inc.
 */
class Page_Paginator
{
    /**
     * アイテムリスト
     * @var array
     */
    private $items;
    /**
     * 基準ページパス
     * @var string
     */
    private $page_path;
    /**
     * ファイル名プレフィックス
     * @var string
     */
    private $prefix;
    /**
     * ファイル名サフィックス
     * @var string
     */
    private $suffix;
    /**
     * １ページあたりのアイテム数
     * @var integer
     */
    private $per_page;
    /**
     * 現在ページ番号
     * @var integer
     */
    private $current_no;
    /**
     * 合計アイテム数
     * @var boolean
     */
    private $item_count;
    /**
     * 合計ページ数
     * @var boolean
     */
    private $page_count;
    /**
     * ページ番号ナビゲーションリンクの数
     * @var integer
     */
    private $mid_size;
    /**
     * ページ番号ナビゲーションの端のリンクの数
     * @var integer
     */
    private $end_size;
    /**
     * ページネーションの現在アイテムクラス名
     * @var string
     */
    private $current_class;
    /**
     * ページネーションのリンクアイテムクラス名
     * @var string
     */
    private $link_class;
    /**
     * ページネーションの省略アイテムクラス名
     * @var string
     */
    private $ellipsis_class;
    /**
     * ページネーションの省略テキスト(...)
     * @var string
     */
    private $ellipsis_text;

    /**
     * コンストラクタ
     * @param array $items
     * @param string $page_path
     * @param string $prefix
     * @param string $suffix
     * @param integer $per_page
     */
    public function __construct($items, $page_path, $prefix = '', $suffix = '', $per_page = 10)
    {
        // 不要な予約IDを削除する
        unset($items['_self_page_id']);

        $this->items = $items;
        $this->page_path = $page_path;
        $this->prefix = $prefix;
        $this->suffix = $suffix;
        $this->per_page = (integer) $per_page;

        // 設定取得
        $this->mid_size = Me_Config::get('mid_size', 'pagination');
        // ページネーションリンクは現在アイテムと前後移動するために最低３つにする。
        if ($this->mid_size < 3) {
            $this->mid_size = 3;
        }
        $this->end_size = Me_Config::get('end_size', 'pagination');
        $this->current_class = Me_Config::get('current_class', 'pagination');
        $this->link_class = Me_Config::get('link_class', 'pagination');
        $this->ellipsis_class = Me_Config::get('ellipsis_class', 'pagination');
        $this->ellipsis_text = Me_Config::get('ellipsis_text', 'pagination');

        // 現在ページは 1 からスタートする
        $this->current_no = 1;
        // 合計アイテム数
        $this->item_count = count($items);
        // 合計ページ数
        $this->page_count = ceil($this->item_count / $this->per_page);
    }

    /**
     * 次のページに進める
     * @return bool
     */
    public function page_feed()
    {
        if ($this->has_next()) {
            $this->current_no++;
            return true;
        } else {
            return false;
        }
    }

    /**
     * ページパス取得
     * @param integer $page_no
     * @return string
     */
    public function get_page_path($page_no = null)
    {
        // ページ番号を省略した場合は現在ページ番号をセットする
        if ($page_no === null) {
            $page_no = $this->current_no;
        }
        // １ページ目は基準ページパスを返す
        if ($page_no == 1) {
            return $this->page_path;
        }
        // ２ページ目以降の処理
        return rtrim(dirname($this->page_path), '/') . '/' . $this->prefix . $page_no . $this->suffix . '.' . pathinfo($this->page_path, PATHINFO_EXTENSION);
    }

    /**
     * 現在ページのアイテムを返す
     * @return array
     */
    public function current_items()
    {
        $current_items = array();

        // 現在ページ用のアイテムのリピート番号を計算
        $repeat_no = (($this->current_no - 1) * $this->per_page) + 1;
        for ($new_repeat_no = 1; $new_repeat_no <= $this->per_page; $new_repeat_no++) {
            if (!isset($this->items[$repeat_no])) {
                break;
            }
            $current_items[$new_repeat_no] = $this->items[$repeat_no++];
        }
        return $current_items;
    }

    /**
     * 予約MEタグ生成
     * @return array
     */
    public function get_reserved()
    {
        $data = array();

        // 合計アイテム数
        $data['_item_count']['node_value'] = $this->item_count;
        // 合計ページ数
        $data['_page_count']['node_value'] = $this->page_count;
        // 現在ページ番号
        $data['_current_no']['node_value'] = $this->current_no;
        // 次へリンク
        $data['_next_url']['node_value'] = $this->next_url();
        // 前へリンク
        $data['_prev_url']['node_value'] = $this->prev_url();
        // 最初のページへのリンク
        $data['_first_url']['node_value'] = $this->first_url();
        // 最後のページへのリンク
        $data['_last_url']['node_value'] = $this->last_url();
        // ページ番号リンクリスト
        $data['_pages'] = $this->get_page_links();

        return $data;
    }

    /**
     * ページ番号リンクリスト取得
     * @return array
     */
    private function get_page_links()
    {
        $pages = array();
        $repeat_no = 1;

        // 開始・終了ページ番号取得
        $start_no = $this->start_no();
        $stop_no = $this->stop_no($start_no);

        // 開始側のページ番号リンク作成
        if ($this->end_size) {
            foreach (range(1, $this->end_size) as $no) {
                if ($no < $start_no) {
                    $this->add_page($pages, $repeat_no++, $no);
                }
            }
            if (($this->end_size + 1) < $start_no) {
                $this->add_page($pages, $repeat_no++, $this->ellipsis_text);
            }
        }

        // ページ番号リンク作成
        foreach (range($start_no, $stop_no) as $no) {
            $this->add_page($pages, $repeat_no++, $no);
        }

        // 終了側のページ番号リンク作成
        if ($this->end_size) {
            if (($this->page_count - $this->end_size) > $stop_no) {
                $this->add_page($pages, $repeat_no++, $this->ellipsis_text);
            }
            foreach (range(($this->page_count - $this->end_size + 1), $this->page_count) as $no) {
                if ($no > $stop_no) {
                    $this->add_page($pages, $repeat_no++, $no);
                }
            }
        }
        return $pages;
    }

    /**
     * ページ番号リンク追加
     * @param array &$pages
     * @param integer $repeat_no
     * @param mixed(integer|string) $no
     */
    private function add_page(&$pages, $repeat_no, $no)
    {
        $pages[$repeat_no]['_no']['node_value'] = $no;
        if (!is_numeric($no)) {
            // ... など数字以外が渡されたばあい
            $pages[$repeat_no]['_url']['node_value'] = '';
            $pages[$repeat_no]['_class']['node_value'] = $this->ellipsis_class;
        } else if ($no === $this->current_no) {
            // 現在ページの場合はURLは返さない
            $pages[$repeat_no]['_url']['node_value'] = '';
            $pages[$repeat_no]['_class']['node_value'] = $this->current_class;
        } else {
            $pages[$repeat_no]['_url']['node_value'] = SITE_PATH . $this->get_page_path($no);
            $pages[$repeat_no]['_class']['node_value'] = $this->link_class;
        }
    }

    /**
     * 前ページの有無を確認
     * @return bool
     */
    private function has_prev()
    {
        return $this->current_no > 1;
    }

    /**
     * 次ページの有無を確認
     * @return bool
     */
    private function has_next()
    {
        return $this->current_no < $this->page_count;
    }

    /**
     * 前ページのURLを返す
     * @return string
     */
    private function prev_url()
    {
        if ($this->has_prev()) {
            return SITE_PATH . $this->get_page_path($this->current_no - 1);
        }
        return '';
    }

    /**
     * 次ページのURLを返す
     * @return string
     */
    private function next_url()
    {
        if ($this->has_next()) {
            return SITE_PATH . $this->get_page_path($this->current_no + 1);
        }
        return '';
    }

    /**
     * 最初ページのURLを返す
     * @return string
     */
    private function first_url()
    {
        if ($this->has_prev()) {
            return SITE_PATH . $this->get_page_path(1);
        }
        return '';
    }

    /**
     * 最後ページのURLを返す
     * @return string
     */
    private function last_url()
    {
        if ($this->has_next()) {
            return SITE_PATH . $this->get_page_path($this->page_count);
        }
        return '';
    }

    /**
     * ページ番号リンクの開始番号を返す
     * @return integer
     */
    private function start_no()
    {
        // 現在ページ番号を基準にして、表示リンク数の半分をオフセットする。
        $no = $this->current_no - floor($this->mid_size / 2);
        // 最終ページに近い場合は開始番号を早める
        if ($no > ($this->page_count - $this->mid_size + 1)) {
            $no = $this->page_count - $this->mid_size + 1;
        }
        // 開始番号が1より小さければ 1 をセットする
        if ($no < 1) {
            $no = 1;
        }
        return (integer) $no;
    }

    /**
     * ページ番号リンクの終了番号を返す
     * @return integer
     */
    private function stop_no($start_no)
    {
        $no = $start_no + $this->mid_size - 1;
        if ($no > $this->page_count) {
            $no = $this->page_count;
        }
        return (integer) $no;
    }
}
